__ALL__ = ["parse_params"]


def parse_params(items):
    params = dict(items)
    for k, v in params.items():
        params[k] = [_.decode() if isinstance(_, bytes) else _ for _ in v]
    return params




import hmac
import tornado.escape
import tornado.web
import tornado.gen
from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor
import mimetypes
import os
import re
import math
import threading
import json
import time
import traceback
import email.utils
import datetime
import omnitools
import utils
import urllib.parse
import requests
from aescipher import AESCipher
from base64 import b32decode, b32encode


__ALL__ = ["BaseResponse", "StaticFileHandler", "ErrorHandler", "NotFound", "File", "HTML", "AJAX"]


class BaseResponse(tornado.web.RequestHandler):
    executor = ThreadPoolExecutor(2**4)
    org_app_root = None
    app_root = None
    sql = None
    writer = None
    cookies_domain = None
    cookies_expires_day = None
    extra_headers = None
    test_server = False
    server = None
    api_key = None
    decrypted_params = None
    export_functions = None
    under_maintenance = False
    internal_error = False
    internal_error_header = None
    internal_error_reason = None
    remote_ip = None
    is_local_ip = False
    is_localhost = False
    is_server_local = False
    admin_contact = ""
    grr_secret = ""

    def prepare(self):
        if self._finished:
            return
        if self.internal_error:
            self.set_status(500)
            try:
                content = '''<html>
<head>
    <meta charset="UTF-8">
    <title>500 Internal Server Error </title>
</head>
<body>
    <h2>{}</h2>
    <div><pre>{}</pre></div>
    <div>Please try again later.</div>{}
</body>
</html>
'''
                content = content.format(
                    (self.internal_error_header or ""),
                    self.format_exc(self.internal_error_reason or ""),
                    self.admin_contact,
                ).encode()
            except:
                traceback.print_exc()
                content = "<title>{msg}</title><body>{msg}{admin_contact}</body>".format(
                    msg="500 Internal Server Error",
                    admin_contact=self.admin_contact
                )
            self.write(content)
            self.finish()
            return
        if self.test_server and not self.is_local_ip:
            self.write_error(403, msg="403 Forbidden")
            return
        if not self.is_localhost:
            try:
                self.prepare_username()
                self.prepare_xsrf()
                self.prepare_session_key()
            except:
                if not self._finished or not self.get_status():
                    self.write_error(500, msg=traceback.format_exc())
                return
            if self.request.headers["Host"] == "127.0.0.1":
                self.write_error(403, msg="403 Forbidden")
                return
            if self.under_maintenance:
                self.set_status(503)
                content = "<title>503 Service Unavailable</title><body><h2>Server Maintenance</h2><div>Please try again later.</div>{}</body>"
                content = content.format(self.admin_contact)
                self.write(content)
                self.finish()
                return
            for k, v in self.extra_headers.items():
                if self.test_server and k == "Content-Security-Policy":
                    continue
                self.set_header(k, v)

    def set_cookie(self, k, v, expires_day=None, **kwargs) -> None:
        if expires_day is None:
            expires_day = self.cookies_expires_day
        if "expires_days" in kwargs:
            kwargs.pop("expires_days")
        if "domain" in kwargs:
            kwargs.pop("domain")
        super().set_cookie(k, v, domain=self.cookies_domain, expires_days=expires_day, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.org_app_root = self.app_root
        if self.server is not None:
            self.app_root = os.path.join(self.app_root, self.server)
        self.cookies_macros = {
            "get": self.get_cookie,
            "set": self.set_cookie,
            "get_secure": self.get_secure_cookie,
            "set_secure": self.set_secure_cookie,
        }
        import string
        if not all(c in string.printable for c in self.request.path):
            try:
                self.request.path = b"".join(bytes([ord(c)]) for c in self.request.path).decode()
            except:
                pass
        if self.request.headers["Host"].startswith("test."):
            self.test_server = True
        self.prepare_request_summary()
        self.is_local_ip = re.search(r"192\.168\.", self.remote_ip) is not None
        self.is_localhost = re.search(r"127\.0\.0\.1", self.remote_ip) is not None

    def prepare_username(self):
        pass

    def prepare_xsrf(self):
        pass

    def prepare_session_key(self):
        pass

    def prepare_request_summary(self):
        try:
            _body = self.request.body.decode()
        except:
            _body = ""
            for _ in self.request.body:
                _body += "\\x{:02x}".format(_)
        headers = dict(self.request.headers)
        self.request_summary = {
            "timestamp": int(time.time())
        }
        if "Host" in headers:
            self.request_summary.update({
                "Host": headers["Host"]
            })
            host = headers.pop("Host")
            self.is_server_local = re.search(r"192\.168\.", host) is not None
        if "X-Forwarded-For" in headers:
            proxied_ip = headers["X-Forwarded-For"].split(",")
            x_real_ip = headers["X-Real-Ip"]
            remote_ip = None
            for ip in proxied_ip:
                ip = ip.strip()
                if re.search(r"^[0-9]{1,3}(\.[0-9]{1,3}){3}$", ip) is not None and ip != x_real_ip:
                    remote_ip = ip
            if not remote_ip:
            #     if len(proxied_ip) > 1:
            #         _index = 1
            #     else:
            #         _index = 0
                remote_ip = proxied_ip[0].strip() or x_real_ip
            self.remote_ip = remote_ip
            headers.pop("X-Real-Ip")
        else:
            self.remote_ip = self.request.remote_ip
        self.request_summary.update({
            "remote_ip": self.remote_ip
        })
        self.request_summary.update({
            "method": self.request.method,
            "uri": self.request.path+("?"+self.request.query if self.request.query else ""),
            "body": _body,
        })
        self.request_summary.update(headers)

    def decrpyt_params(self, v):
        return v

    def encrpyt_params(self, v):
        return v

    def check_xsrf_cookie(self):
        raise NotImplementedError()

    def dump_request_summary(self, fn):
        self.writer(
            os.path.join(self.app_root, "log", fn+".log"),
            "ab",
            (json.dumps(self.request_summary, ensure_ascii=False)+"\n").encode()
        )

    def check_etag_header(self):
        return False

    def compute_etag(self):
        return None

    def format_exc(self, exc):
        exc = re.sub(r"File.*?site-packages.", "File \"", exc)
        regex = re.compile("File.*?{}.".format(re.escape(self.org_app_root)))
        exc = regex.sub("File \"app_root\\\\", exc)
        return exc

    def write_error(self, status_code, **kwargs):
        self.dump_request_summary("error")
        msg = {
            "status_code": status_code,
            "msg": "unknown error" if "msg" not in kwargs else self.format_exc(kwargs["msg"])
        }
        self.set_header("Content-Type", "text/plain")
        if status_code == 406:
            na_406_html = "<html><title>406 Not Acceptable</title><body><h1>{}</h1><script>setTimeout(function(){{window.location.reload(true)}}, 5000);</script></body></html>"
            self.set_status(200)
            self.set_header("Content-Type", "text/html")
            self.write(na_406_html.format("IP Error\n(Please do not use company network or public wi-fi)\nRefresh to retry."))
        else:
            self.set_status(status_code)
            self.write(json.dumps(msg))
        self.finish()


class StaticFileHandler(BaseResponse, tornado.web.StaticFileHandler):
    pass


class ErrorHandler(BaseResponse):
    def check_xsrf_cookie(self):
        return


class NotFound(ErrorHandler):
    def prepare(self):
        self.write_error(404, msg="File Not Found")


class SuccessHandler(BaseResponse):
    rh = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        import time
        # self.ts = self.request.remote_ip+"/"+str(time.time())+"/"+self.request.path
        # SuccessHandler.rh.append(self.ts)
        p = threading.Thread(target=self._monitor_count)
        p.daemon = True
        p.start()

    def _monitor_count(self):
        while True:
            if self._finished:
                # SuccessHandler.rh.remove(self.ts)
                break
            import time
            # time.sleep(1/1000)
            time.sleep(1)
        self.dump_request_summary("success")


class BaseFile(SuccessHandler):
    _args = None
    file_name = None
    file_path = None
    file_modified = None

    def get_file_stat(self):
        return os.stat(self.file_path)

    def get_file_modified(self):
        return datetime.datetime.utcfromtimestamp(int(self.get_file_stat().st_mtime))

    def should_return_304(self) -> bool:
        ims_value = self.request.headers.get("If-Modified-Since")
        if ims_value is not None:
            date_tuple = email.utils.parsedate(ims_value)
            if date_tuple is not None:
                if_since = datetime.datetime(*date_tuple[:6])
                if if_since >= self.file_modified:
                    return True
        return False

    def set_modified_header(self):
        self.set_header("Last-Modified", self.file_modified)

    def set_304_header(self):
        self.set_status(304)

    def get_what(self, og_title=None, og_description=None):
        return None

    @run_on_executor
    def _get(self):
        self.set_modified_header()
        if self.should_return_304():
            return self.set_304_header()
        what = self.get_what() or open(self.file_path, "rb").read()
        if "Content-Security-Policy" in self._headers:
            self._headers["Content-Security-Policy"] = self._headers["Content-Security-Policy"].replace(" 'nonce-foxe6'", "")
        return what

    @tornado.gen.coroutine
    def get(self, *args):
        self._args = args
        a = yield self._get()
        if a:
            self.write(a)
            self.finish()


class File(BaseFile):
    def prepare(self) -> None:
        super().prepare()
        self.file_path = os.path.join(self.app_root, self.path_args[0])
        content_type = mimetypes.guess_type(self.file_path)[0]
        self.set_header("Content-Type", content_type or "application/octet-stream")
        self.file_modified = self.get_file_modified()


class HTML(BaseFile):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.file_path = os.path.join(self.app_root, self.file_name+".html")
        self.file_modified = self.get_file_modified()


class AJAX(SuccessHandler):
    api_name = None

    @run_on_executor
    def _run(self, method: str):
        params = self.decrypted_params if self.decrypted_params else utils.parse_params(self.request.arguments.items())
        exec(open(os.path.join(self.app_root, self.api_name + "_api.py"), "rb").read().decode())
        return locals()["api_"+method](
            params=params,
            cookies=self.cookies_macros,
            app_root=self.app_root,
            sql=self.sql,
            writer=self.writer,
            export_functions=self.export_functions,
            x_real_ip=self.remote_ip,
            test_server=self.test_server,
            domain=self.cookies_domain[1:],
            server_name=self.server
        )

    @tornado.gen.coroutine
    def get(self, *args):
        raw_params = yield self._run("get")
        self.write(raw_params)
        self.finish()

    @tornado.gen.coroutine
    def post(self, *args):
        raw_params = yield self._run("post")
        is_exc = isinstance(raw_params, Exception)
        is_traceback = False
        if not is_exc:
            is_traceback = "Traceback (most recent call last):" in raw_params
        if is_exc or is_traceback:
            if is_traceback:
                raw_params = json.loads(raw_params)
            if is_exc:
                raw_params = str(raw_params)
            return self.write_error(400, msg=raw_params)
        raw_params = raw_params.encode()
        if self.decrypted_params:
            self.write(self.encrpyt_params(raw_params))
        else:
            self.set_header("Content-Type", "application/json")
            self.write(raw_params)
        self.finish()





from dwa import *
import pages
import jobs


def app_workers() -> tuple:
    return (
        jobs.shop.item_info_worker(),
        jobs.shop.reserve_order_worker(),
    )


def get_admin_app() -> tuple:
    return (
        "127.0.0.1",
        pages.shop.get_shop_admin_settings()
    )


def app_settings_template(app_root: str, domain: str, port: int, cookies_expires_day: float) -> dict:
    def csp():
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Security-Policy
        # https://content-security-policy.com/
        csp = {
            "base-uri": [
                "'self'",
            ],
            "connect-src": [
                "'self'",
            ],
            "default-src": [
                "'none'",
            ],
            "font-src": [
                "'none'",
            ],
            "form-action": [
                "'self'",
            ],
            "frame-ancestors": [
                "'none'",
            ],
            "frame-src": [
                "https://www.google.com",
            ],
            "img-src": [
                "*",
                "data:",
            ],
            "script-src": [
                "'self'",
                "https://cdnjs.cloudflare.com",
            ],
            "style-src": [
                "'self'",
                "'unsafe-inline'",
            ],
            "worker-src": [
                "'none'",
            ]
        }
        return "; ".join(["{} {}".format(k, " ".join(v)) for k, v in csp.items()])
    def sql_watcher(*args, **kwargs):
        # monitor sql
        # if validated:
        return True
        # else:
        #     raise Exception("something")
    def raise_backend_errors(header, reason):
        # notify via apprise
        pass
    settings = {}
    db = "db/db.db"
    settings["cookie_secret"] = open(os.path.join(app_root, "cookie_secret.txt"), "rb").read()
    settings["port"] = port
    settings["db"] = db
    settings["db_port"] = 39292+0
    settings["writer_port"] = 39292+1
    settings["domain"] = domain
    settings["cookies_expires_day"] = cookies_expires_day
    settings["extra_headers"] = {
        "Referrer-Policy": "no-referrer",
        "Cache-Control": "must-revalidate, max-age=0",
        "Content-Security-Policy": csp(),
    }
    settings["sql_watcher"] = sql_watcher
    settings["raise_backend_errors"] = raise_backend_errors
    settings["admin_contact"] = "<div>Admin contact: test@example.com</div>"
    settings["grr_secret"] = "google recaptcha secret key"
    servers = [
        pages.root.get_root_settings(),
        pages.shop.get_shop_settings(),
        get_admin_app(),
    ]
    settings["servers"] = {k: v for k, v in servers}
    return settings


def app_settings(app_root: str) -> dict:
    return app_settings_template(app_root, "localtest.me", 8888, 0.25)





import tornado.web
import tornado.ioloop
import tornado.platform.asyncio
import asyncio
import threading
import time
import omnitools
import sqlq
import re
import handlers


__ALL__ = ["TA"]


asyncio.set_event_loop_policy(tornado.platform.asyncio.AnyThreadEventLoopPolicy())


class TA(object):
    sqlqueue = None
    writer = None

    def __init__(
            self, domain: str, servers: dict,
            db: str, db_port: int, writer_port: int,
            cookie_secret: str,
            port: int = 8888,
            xsrf_cookies: bool = True, compress_response: bool = True,
            export_functions = None
    ):
        def _translate_host(server, pages, i):
            NotFound_page = handlers.NotFound
            if server == "" or server == "root":
                host = domain
            elif re.search(r"[0-9]{1,3}(\.[0-9]{1,3}){3}", server):
                host = server
            elif server == "*":
                host = ".*\.?"+domain
            else:
                NotFound_page = type("NotFound_{}".format(server), (handlers.NotFound,), {"server": server})
                host = server+"."+domain
            print("Server {}.1: {}".format(i+1, host))
            servers = [(tornado.web.HostMatches(host), pages+[(r"/(.*)", NotFound_page)])]
            if host != server and server != "*":
                print("Server {}.2: {}".format(i+1, "test."+host))
                File_page = type("File_test_{}".format(server), (handlers.File,), {"server": server})
                servers.append((tornado.web.HostMatches("test."+host), pages+[(r"/(.*\.(?:html|css|js|png|jpeg|jpg|svg|bmp))", File_page)]))
            return servers
        _handlers = []
        for i, (k, v) in enumerate(servers.items()):
            _handlers += _translate_host(k, v, i)
        _handlers.append((tornado.web.HostMatches(r".*"), [(r"/(.*)", handlers.NotFound)]))
        self.app = tornado.web.Application(
            _handlers,
            cookie_secret=cookie_secret,
            xsrf_cookies=xsrf_cookies,
            compress_response=compress_response,
        )
        self.port = port
        if db_port > 0:
            self.sqlqueue = sqlq.SqlQueueU(server=True, db=db, db_port=db_port, timeout_commit=30*1000, auto_backup=False, export_functions=export_functions)
        if writer_port > 0:
            self.writer = omnitools.WriterU(server=True, writer_port=writer_port)

    def _start(self) -> None:
        self.server = self.app.listen(self.port)
        self.server_loop = tornado.ioloop.IOLoop.instance()
        self.server_loop.start()

    def start(self) -> None:
        p = threading.Thread(target=self._start)
        p.daemon = True
        p.start()

    def stop(self, stop_app_worker) -> None:
        self.server_loop.stop()
        self.server.stop()
        while True:
            if len(handlers.SuccessHandler.rh) == 0:
                break
            print("no. of requests remaining:", len(handlers.SuccessHandler.rh), flush=True)
            time.sleep(0.5)
        stop_app_worker()
        if self.sqlqueue:
            self.sqlqueue.commit()
            self.sqlqueue.stop()
        if self.writer:
            self.writer.stop()





__version__ = "0.0.36"
__keywords__ = ["tornado ajax wrapper framework"]


# if not __version__.endswith(".0"):
#     import re
#     print("version {} is deployed for automatic commitments only".format(__version__), flush=True)
#     print("install version " + re.sub(r"([0-9]+\.[0-9]+\.)[0-9]+", r"\g<1>0", __version__) + " instead")
#     import os
#     os._exit(1)


import os
import sys
sys.path.extend([os.path.dirname(os.path.abspath(__file__))])


import threading
import time
import re
import unicodedata
import traceback
import tornado
import tornado.web
import tornado.gen
import tornado.escape
import tornado.concurrent
import tornado.template
import core
import handlers
import workers
import requests
import aescipher
import sqlq
import unencryptedsocket
import threadwrapper
import omnitools
import lxml



class DWA(object):
    app_root = None
    app_settings = None
    app_workers = None
    ta = None

    def __init__(self, domain=None, port=None, cookies_expires_day=None):
        self.port = port
        self.domain = domain
        self.cookies_expires_day = cookies_expires_day
        self.clear()
        self.app_root = omnitools.abs_main_dir(2)
        sys.path.append(self.app_root)
        print("app root:", self.app_root)
        self.setup()

    def start_app_worker(self, raise_backend_errors):
        tw = threadwrapper.ThreadWrapper(threading.Semaphore(2**3))
        for app_worker in self.app_workers:
            tw.add(app_worker.start)
        tw.wait()
        p = threading.Thread(target=self.check_app_worker, args=(raise_backend_errors,))
        p.daemon = True
        p.start()

    def stop_app_worker(self):
        tw = threadwrapper.ThreadWrapper(threading.Semaphore(2**3))
        for app_worker in self.app_workers:
            tw.add(app_worker.stop)
        tw.wait()

    def check_app_worker(self, raise_backend_errors):
        def notify(app_worker):
            header = "Worker {} Exception".format(app_worker.__class__.__name__)
            reason = app_worker.exploded_reason
            handlers.BaseResponse.internal_error = True
            handlers.BaseResponse.internal_error_header = header
            handlers.BaseResponse.internal_error_reason = reason
            try:
                raise_backend_errors(header, reason)
            except:
                print(header)
                print(reason)
        stop = False
        while not stop:
            for app_worker in self.app_workers:
                if app_worker.exploded:
                    stop = True
                    notify(app_worker)
            if self.ta.sqlqueue:
                if self.ta.sqlqueue.ioerror:
                    stop = True
                    notify(self.ta.sqlqueue)
            if stop:
                break
            time.sleep(1)

    def setup(self):
        exec(open(os.path.join(self.app_root, "app.py"), "rb").read().decode(), globals())
        self.app_settings = globals()["app_settings"](self.app_root)
        self.app_settings["port"] = self.port or self.app_settings["port"]
        self.app_settings["domain"] = self.domain or self.app_settings["domain"]
        self.app_settings["cookies_expires_day"] = self.cookies_expires_day or self.app_settings["cookies_expires_day"]
        self.app_workers = globals()["app_workers"]()
        raise_backend_errors = self.app_settings["raise_backend_errors"]
        handlers.BaseResponse.app_root = self.app_root
        workers.base_worker.app_root = self.app_root
        handlers.BaseResponse.port = self.app_settings["port"]
        handlers.BaseResponse.cookies_domain = "." + self.app_settings["domain"]
        handlers.BaseResponse.cookies_expires_day = self.app_settings["cookies_expires_day"]
        handlers.BaseResponse.extra_headers = self.app_settings["extra_headers"]
        handlers.BaseResponse.under_maintenance = True
        handlers.BaseResponse.admin_contact = self.app_settings["admin_contact"]
        handlers.BaseResponse.grr_secret = self.app_settings["grr_secret"]
        workers.base_worker.under_maintenance = True
        if not os.path.isabs(self.app_settings["db"]):
            self.app_settings["db"] = os.path.join(self.app_root, self.app_settings["db"])
        def export_functions(k, *args, **kwargs):
            for app_worker in self.app_workers:
                f = app_worker.export_functions or {}
                if k in f:
                    try:
                        return f[k](*args, **kwargs)
                    except:
                        return traceback.format_exc()
            raise Exception("no function named: {}".format(k))
        class_export_functions = staticmethod(export_functions)
        self.ta = core.TA(
            self.app_settings["domain"],
            self.app_settings["servers"],
            self.app_settings["db"],
            self.app_settings["db_port"],
            self.app_settings["writer_port"],
            omnitools.sha3_512hd(self.app_settings["cookie_secret"]),
            self.app_settings["port"],
            export_functions=export_functions
        )
        @staticmethod
        def sql(*args, **kwargs):
            if self.ta.sqlqueue:
                result = self.app_settings["sql_watcher"](*args, sql=self.ta.sqlqueue.sql, **kwargs)
                if result != True:
                    if isinstance(result, Exception):
                        raise result
                    else:
                        raise Exception(result)
                result = self.ta.sqlqueue.sql(*args, **kwargs)
                if isinstance(result, str):
                    raise Exception(result)
                elif isinstance(result, Exception):
                    raise result
                return result
            else:
                return False
        @staticmethod
        def writer(*args, **kwargs):
            if self.ta.writer:
                return self.ta.writer.write(*args, **kwargs)
            else:
                return False
        handlers.BaseResponse.sql = sql
        handlers.BaseResponse.writer = writer
        workers.base_worker.sql = sql
        workers.base_worker.writer = writer
        handlers.BaseResponse.export_functions = class_export_functions
        self.ta.start()
        self.start_app_worker(raise_backend_errors)
        print("[server] started")
        self.interactive_input()

    @staticmethod
    def printTable(result, columns=None):
        def _len(s):
            length = 0
            for c in s:
                status = unicodedata.east_asian_width(c)
                if status in ["W", "F", "A"]:
                    length += 2
                else:
                    length += 1
            return length
        if not columns:
            columns = list(result[0].keys() if result else [])
        table = [columns]
        for row in result:
            table.append([str(row[column] if row[column] is not None else "") for column in columns])
        column_width = [max(map(_len,column)) for column in zip(*table)]
        table.insert(1, ["-" * i for i in column_width])
        for row in table:
            print("|".join([" {}{} ".format(column, " "*(column_width[i]-_len(column))) for i, column in enumerate(row)]))

    @staticmethod
    def clear():
        if sys.platform == "win32":
            os.system("cls")
        else:
            os.system("clear")

    def interactive_input(self):
        while True:
            print()
            print("Enter 'help' for commands.")
            command = input("> ")
            if not command:
                continue
            elif command == "help":
                print('''Commands:
        backup: Backup sqlite database
        commit: Commit sqlite database
        maintenance: Render server as '503 Service Unavailable'
        resume: Resume server from maintenance
        sql <SQL statement>: Execute SQL statement
        stop: Stop server and commit
        terminate: Force stop server''')
            elif command == "backup":
                self.ta.sqlqueue.backup()
                print("[sqlite] backup")
            elif command == "clear":
                self.clear()
            elif command == "commit":
                self.ta.sqlqueue.commit()
                print("[sqlite] committed")
            elif command == "maintenance":
                handlers.BaseResponse.under_maintenance = True
                workers.base_worker.under_maintenance = True
                print("[maintenance] True")
            elif command == "resume":
                handlers.BaseResponse.under_maintenance = False
                workers.base_worker.under_maintenance = False
                print("[maintenance] False")
            elif command.startswith("sql"):
                sql = command[4:]
                try:
                    while not sql.endswith(";"):
                        sql += " "+input("> ... ")
                except KeyboardInterrupt:
                    print()
                    continue
                data = []
                if "?" in sql:
                    last = sql.count("?")
                    for i in range(0, last):
                        _data = input("> data ({}/{}) ".format(i+1, last))
                        if _data.isdigit():
                            _data = int(_data)
                        data.append(_data)
                try:
                    result = self.ta.sqlqueue.sql(sql, tuple(data))
                    self.printTable(result)
                except:
                    pass
            elif command == "stop":
                print("[server] stopping")
                self.ta.stop(lambda : self.stop_app_worker())
                print("[server] exiting")
                time.sleep(2)
                return
            elif command == "terminate":
                print("[server] terminating")
                return
            else:
                print("! unknown command")
            print()


